import collections

from . import ql


class MapperMetaclass(type):

    @classmethod
    def __prepare__(metacls, name, bases, **kwds):
        return collections.OrderedDict()

    def __new__(cls, name, bases, namespace, **kwds):
        result = type.__new__(cls, name, bases, dict(namespace))
        result.fields = []
        # Copy fields from bases
        for base in bases:
            if hasattr(base, 'fields'):
                for field in base.fields:
                    field.contribute_to_mapper(result, field.name)
        for field_name, field in namespace.items():
            if hasattr(field, 'contribute_to_mapper'):
                field.contribute_to_mapper(result, field_name)
        return result


class Mapper(object, metaclass=MapperMetaclass):

    def __init__(self, fields=ql.every_dict):
        if isinstance(fields, str):
            fields = ql.parse(fields)
        self.include_fields = fields
        self.fields = [
            field.bind(self)
            for field in self.fields
            if field.name in fields
        ]

    def dump(self, obj, target=None):
        if target is None:
            target = {}
        for field in self.fields:
            field.dump(obj, target)
        return target

    def drf(self):
        '''
        Return a mapper that is compatible to the Django REST Framework
        serializer API.
        '''
        from lion.contrib.drf import DRFMapper
        return DRFMapper.wrap(self)



from pyparsing import alphas, alphanums, delimitedList, Dict, Forward, \
        Group, Literal, Optional, Suppress, Word

bracket_open = Suppress(Literal('{'))
bracket_close = Suppress(Literal('}'))
identifier = Word(alphas, alphanums + "_" ).setName("identifier")
field = Forward()
fields = delimitedList(field)
obj = Dict(bracket_open + fields + bracket_close)
field << Group(identifier + Optional(obj, default=None))

ql = obj

class EveryDict(object):

    def __contains__(self, name):
        return True

    def __getitem__(self, name):
        return self

every_dict = EveryDict()

def to_dict(d):
    return {
        k: to_dict(v) if v != None else every_dict
        for k, v in d.items()
    }

def parse(s):
    if not s:
        return every_dict
    return to_dict(ql.parseString(s))


from copy import copy, deepcopy

from .conditions import no_condition


class BaseField(object):

    def dump(self, obj, target):
        raise NotImplemented("Subclasses of BaseField need to implement this method.")

    def bind(self, mapper):
        # Most fields don't need anything special when being bound so
        # they simply return theirselves.
        return self


class Field(BaseField):

    def __init__(self, source=None, getter=None, condition=no_condition):
        self.name = None
        self.source = source
        if getter:
            self.getter = getter
        self.condition = condition

    def getter(self, obj, name):
        return getattr(obj, name)

    def dump(self, obj, target):
        value = self.getter(obj, self.source)
        if self.condition(value):
            value = self.denormalize_value(value)
            target[self.name] = value

    def denormalize_value(self, value):
        return value

    def contribute_to_mapper(self, mapper, name):
        clone = deepcopy(self)
        clone.name = name
        clone.source = clone.source or name
        mapper.fields.append(clone)


class BoolField(Field):

    def denormalize_value(self, value):
        if value is None:
            return None
        return bool(value)


class IntField(Field):

    def denormalize_value(self, value):
        if value is None:
            return None
        return int(value)


class FloatField(Field):

    def denormalize_value(self, value):
        if value is None:
            return None
        return float(value)


class StrField(Field):

    def denormalize_value(self, value):
        if value is None:
            return None
        return str(value)


class ConstField(BaseField):

    def __init__(self, value):
        self.value = value

    def dump(self, obj, target):
        target[self.name] = self.value

    def contribute_to_mapper(self, mapper, name):
        clone = deepcopy(self)
        clone.name = name
        mapper.fields.append(clone)


class UUIDField(StrField):
    pass


class DateTimeField(Field):

    def __init__(self, source=None, getter=None, condition=no_condition, tz=None):
        super().__init__(source, getter, condition)
        self.tz = tz

    def denormalize_value(self, value):
        if value is None:
            return None
        if self.tz is not None:
            if value.tzinfo is not None:
                value = value.astimezone(self.tz)
            else:
                value = self.tz.localize(value)
        return value.isoformat()


class MapperMethodField(BaseField):

    def __init__(self, method_name=None, condition=no_condition):
        self.name = None
        self.method_name = method_name
        self.condition = condition
        self.mapper = None

    def bind(self, mapper):
        clone = copy(self)
        clone.mapper = mapper
        return clone

    def dump(self, obj, target):
        method = getattr(self.mapper, self.method_name)
        value = method(obj)
        if self.condition(value):
            target[self.name] = value

    def contribute_to_mapper(self, mapper, name):
        clone = deepcopy(self)
        clone.mapper = mapper
        clone.name = name
        clone.method_name = self.method_name or 'get_' + name
        mapper.fields.append(clone)


class ListField(Field):

    def __init__(self, mapper=None, **kwargs):
        super().__init__(**kwargs)
        self.mapper_class = mapper
        self.mapper = None
        self.kwargs = kwargs

    def bind(self, mapper):
        clone = copy(self)
        clone.mapper = clone.mapper_class(mapper.include_fields[self.name])
        return clone

    def denormalize_value(self, value):
        if value is None:
            return None
        mapper = self.mapper
        if isinstance(mapper, BaseField):
            return [
                mapper.denormalize_value(x)
                for x in value
            ]
        else:
            return [
                mapper.dump(x)
                for x in value
            ]


class MapperField(Field):

    def __init__(self, mapper, **kwargs):
        super().__init__(**kwargs)
        self.mapper_class = mapper
        self.mapper = None

    def bind(self, mapper):
        clone = copy(self)
        clone.mapper = clone.mapper_class(mapper.include_fields[self.name])
        return clone

    def denormalize_value(self, value):
         if value is None:
             return None
         return self.mapper.dump(value)



def update_object(obj, data, fields):
    '''
    Update an object by using the values from `data` limited by
    the fields in `fields`. Returns `True` if the object was modified.
    '''
    modified = False
    for field in fields:
        if field in data:
            # FIXME smart conversion of data types
            value = data[field]
            if getattr(obj, field) != value:
                setattr(obj, field, data[field])
                modified = True
    return modified


def no_condition(value):
    return True

def skip_none(value):
    return value is not None

def skip_empty(value):
    return value is not None and len(value) > 0

def skip_false(value):
    return bool(value)



from .mapper import Mapper
from .fields import Field, BoolField, IntField, FloatField, StrField, \
        ConstField, UUIDField, DateTimeField, MapperMethodField, \
        ListField, MapperField
from .update import update_object
from .conditions import no_condition, skip_none, skip_empty, skip_false
from . import getters
from .ql import parse as parse_ql

__version__ = '0.1.0'



from functools import partial

class DRFMapper(object):
    '''
    Wrapper class to support the Serializer API of the Django REST Framework.
    Call `wrap(mapper)` or `mapper.drf()` to get a useable instance of this.
    '''

    def __init__(self, obj, many=False):
        self.obj = obj
        self.many = many

    @classmethod
    def wrap(cls, mapper):
        class BoundDRFMapper(DRFMapper):
            pass
        BoundDRFMapper.mapper = mapper
        return BoundDRFMapper

    @property
    def data(self):
        if self.many:
            return [self.mapper.dump(x) for x in self.obj]
        else:
            return self.mapper.dump(self.obj)



#!/usr/bin/env python

import os
import sys

from setuptools import setup

if sys.argv[-1] == 'publish':
    os.system("python setup.py sdist upload")
    os.system("python setup.py bdist_wheel upload")
    print("You probably want to also tag the version now:")
    print("  python setup.py tag")
    sys.exit()

if sys.argv[-1] == 'tag':
    os.system("git tag -a v%s -m 'Version %s'" % (version, version))
    os.system("git push --tags")
    sys.exit()

setup()
